(in-package :empire)

(defparameter *version* "0.0.1")

;;
;; Connection
;;

(defvar *default-empire-server* "localhost")
(defvar *default-empire-server-port* 6665)

(defclass connection ()
  ((user
    :initarg :user
    :accessor user)
   (password
    :initarg :password
    :accessor password
    :initform nil)
   (server-name
    :initarg :server-name
    :accessor server-name
    :initform "Unknown server")
   (server-port
    :initarg :server-port
    :accessor server-port
    :initform *default-empire-server-port*)
   (socket
    :initarg :socket
    :accessor socket
    :documentation "Slot to store socket (for internal use only).")
   (stream
    :initarg :stream
    :accessor network-stream
    :documentation "Stream used to talk to the empire server.")
   (connection-handler
    :accessor connection-handler
    :initform nil
    :documentation "Handler function for incoming messages")
   (mode
    :accessor connection-mode
    :initform nil
    :documentation "TODO(gefla)")
   (session
    :accessor session
    :initarg :session
    :documentation "web session to which this connection belongs")
   (send-queue
    :accessor send-queue
    :initform (locked-queue:create)
    :documentation "lines outstanding to be sent at next prompt")))

(defun make-connection (&key (user nil)
			(password nil)
			(server-name *default-empire-server*)
			(server-port *default-empire-server-port*)
			(socket nil)
			(network-stream nil)
			session)
  (make-instance 'connection
		 :user user
		 :password password
		 :server-name server-name
		 :server-port server-port
		 :socket socket
		 :stream network-stream
		 :session session))

;;; interface
(defun connect  (&key (user nil)
		 (password nil)
		 (server-name *default-empire-server*)
		 (server-port *default-empire-server-port*)
		 session)
  "Connect to server and return a connection object."
  (let* ((socket (usocket:socket-connect server-name server-port))
         (stream (usocket:socket-stream socket))
         (connection (make-connection :user user
				      :password password
				      :socket socket
                                      :network-stream stream
                                      :server-name server-name
				      :server-port server-port
				      :session session))
	 (handler (spawn-with-name (format nil "empire-handler-~a" user)
				   #'handle-connection connection)))
    (setf (connection-handler connection) handler)
    (empire-log:info "~a: connect" connection)
    connection))

(defgeneric quit (connection))
(defgeneric handle-connection (connection))
(defgeneric read-message (connection))
(defgeneric send-message (connection message &key next-mode))
(defgeneric connected-p (connection))
(defgeneric reconnect (conncetion))

(defmethod connected-p ((c connection))
  (sb-thread:thread-alive-p (connection-handler c)))

(defmethod reconnect ((c connection))
  (with-slots (server-name server-port user password socket stream connection-handler) c
      (let* ((sock (usocket:socket-connect server-name server-port))
	     (s (usocket:socket-stream sock)))
	(setf socket sock
	      stream s
	      connection-handler (spawn-with-name (format nil "empire-handler-~a" user)
						  #'handle-connection c))))
  (empire-log:info "~a: reconnect" c))

(defparameter +C_CMDOK+ "0")
(defparameter +C_DATA+ "1")
(defparameter +C_INIT+ "2")
(defparameter +C_EXIT+ "3")
(defparameter +C_FLUSH+ "4")
(defparameter +C_NOECHO+ "5")
(defparameter +C_PROMPT+ "6")
(defparameter +C_ABORT+ "7")
(defparameter +C_REDIR+ "8")
(defparameter +C_PIPE+ "9")
(defparameter +C_CMDERR+ "a")
(defparameter +C_BADCMD+ "b")
(defparameter +C_EXECUTE+ "c")
(defparameter +C_FLASH+ "d")
(defparameter +C_INFORM+ "e")

(defclass base-mode ()
  ((connection :initarg :connection :accessor connection)))
(defclass init-mode (base-mode)
  ((play-sent-p :initform nil
		:documentation "Have we already sent the play command?")))
(defclass play-mode (base-mode) nil)

(defun make-mode (connection mode)
  (make-instance mode :connection connection))

(defgeneric set-new-mode (base-mode new-mode))
(defmethod set-new-mode ((m base-mode) new-mode)
  (let* ((c (connection m))
	 (mode (make-mode c new-mode)))
    (empire-log:info "~a: set-new-mode ~a -> ~a" c (connection-mode c) new-mode)
    (setf (connection-mode c) mode)))

(defgeneric handle-data (base-mode message))
(defgeneric handle-init (base-mode message))
(defgeneric handle-exit (base-mode message))
(defgeneric handle-prompt (base-mode message))
(defgeneric handle-flush (base-mode message))
(defgeneric handle-simple-message (base-mode message))
(defgeneric handle-ignore (base-mode message))

(defmethod handle-init ((m init-mode) message)
  (declare (ignorable message))
  (with-slots (play-sent-p) m
    (if (not play-sent-p)
	(let ((c (connection m)))
	  (login-and-play c)
	  (setf play-sent-p t))
	(set-new-mode m 'play-mode))))

(defmethod handle-exit ((m init-mode) message)
  (declare (ignorable message))
  t)

(defmethod handle-exit ((m play-mode) message)
  (declare (ignorable message))
  (quit (connection m)))

(defmethod handle-data ((m play-mode) message)
  (let* ((c (connection m)))
    (empire-web:data (session c) message)))

(defmethod handle-simple-message ((m play-mode) message)
  (let* ((c (connection m)))
    (empire-web:data (session c) message)))

(defmethod handle-flush ((m play-mode) message)
  (let* ((c (connection m)))
    (empire-web:prompt (session c) message)
    (send-next-line c)))

(defmethod send-next-line ((c connection))
  (let ((next-event (locked-queue:dequeue (send-queue c))))
    (etypecase next-event
      (string (send-message-one c next-event))
      (cons (let ((message (car next-event))
		  (mode (cdr next-event)))
	      (send-message-one c message)
	      (set-new-mode c mode))))))

(defmethod handle-prompt ((m play-mode) message)
  (let* ((c (connection m)))
    (with-input-from-string (s message)
      (let* ((minutes (read s)) ;;FIXME DANGEROUS
	     (btus (read s)))
	(empire-web:prompt (session c)
			   (format nil "[~a,~a]: " minutes btus))))
    (send-next-line c)))

(defmethod handle-ignore ((m init-mode) message)
  (declare (ignorable message))
  t)

(defparameter *line-type-dispatch*
  (list `(,+C_DATA+ . handle-data)
	`(,+C_INIT+ . handle-init)
	`(,+C_EXIT+ . handle-exit)
	`(,+C_FLUSH+ . handle-flush)
	`(,+C_BADCMD+ . handle-simple-message)
	`(,+C_FLASH+ . handle-simple-message)
	`(,+C_CMDOK+ . handle-ignore)
	`(,+C_PROMPT+ . handle-prompt)))

(defun parse-server-line (line)
  (let* ((first-space-index
	  (loop for i from 0 below (length line)
	     if (eq (char line i) #\Space)
	     do (return i)))
	 (message-type (subseq line 0 first-space-index))
	 (message (if first-space-index
		      (subseq line (+ first-space-index 1))
		      "")))
    (values message message-type)))

(defmethod handle-connection ((c connection))
  (setf (connection-mode c) (make-mode c 'init-mode))
  (loop while (read-message c)))

(defun lookup-handler (message-type)
  (cdr (assoc message-type *line-type-dispatch* :test 'string=)))

(define-condition no-handler (error)
  ((mode :initarg :mode)
   (type :initarg :type))
  (:report (lambda (condition stream)
	     (with-slots (mode type) condition
	       (format stream "No handler for ~a in mode ~a."
		       type mode)))))
(defmethod read-message ((c connection))
  (handler-case
      (let* ((s (network-stream c))
	     (line (read-line s)))
	(empire-log:info "~a: < ~a" c line)
	(multiple-value-bind (message type) (parse-server-line line)
	  (let ((handler (lookup-handler type))
		(mode (connection-mode c)))
	    (if handler
		(funcall handler mode message)
		(error 'no-handler :mode mode :type type))))
	line)
    (sb-int:closed-stream-error () nil)
    (end-of-file () nil)))

(defun raw-send-message (s message)
    (write-sequence message s)
    (terpri s)
    (force-output s))

(defgeneric send-message-one (connection string))
(defmethod send-message-one ((c connection) message)
  (empire-log:info "~a: > ~a" c message)
  (let ((s (network-stream c)))
    (raw-send-message s message)))

(defmethod send-message ((c connection) message &key next-mode)
  (if (not (connected-p c))
      (reconnect c))
  (if next-mode
      (locked-queue:enqueue (send-queue c) (cons message next-mode))
      (locked-queue:enqueue (send-queue c) message)))

(defparameter +special-command-char+ #\;)

(defun special-command-p (line)
  (and (> (length line) 0)
       (char= +special-command-char+ (aref line 0))))

(defmethod special-xup ((c connection))
  t)

(defmethod special-command ((c connection) line)
  (cond ((string= line "xup") (special-xup c))
	(t (empire-web:data (session c) "Unknown special command")))
  line)

(defmethod command ((c connection) line)
  (cond ((special-command-p line) (special-command c (subseq line 1)))
	(t (send-message c line))))

(defgeneric login-and-play (connection))
(defmethod login-and-play ((c connection))
  (send-message-one c (format nil "client eow ~a" *version*))
  (send-message-one c "user FIXME")
  (with-slots (user password) c
    (send-message-one c (format nil "coun ~a" user))
    (send-message-one c (format nil "pass ~a" password)))
  (send-message-one c "kill")
  (send-message-one c "play"))

(defmethod quit ((c connection))
  (usocket:socket-close (socket c)))


;;; Tests
(defvar *l1* "2 Empire server ready")
;; (parse-server-line *l1*)
